package org.feup.apm.myauthenticator

import android.content.Context
import android.content.SharedPreferences
import android.icu.util.GregorianCalendar
import android.security.keystore.KeyGenParameterSpec
import android.security.keystore.KeyProperties
import android.util.Base64
import androidx.biometric.BiometricPrompt
import androidx.security.crypto.EncryptedSharedPreferences
import androidx.security.crypto.MasterKeys
import java.math.BigInteger
import java.security.KeyPairGenerator
import java.security.KeyStore
import java.security.Signature
import java.security.spec.ECGenParameterSpec
import javax.security.auth.x500.X500Principal

object Crypto {
  const val PROVIDER = "AndroidKeyStore"
  const val AUTH_KEY_ALIAS = "authenticator_key"
  const val KEY_ALGORITHM = "EC"
  const val CERT_SERIAL = 45124963L
  const val SIGN_ALGORITHM = "SHA256withECDSA"
}

fun byteArraytoHex(byteArray: ByteArray): String {
  return byteArray.joinToString(":") { "%02x".format(it) }
}

fun generateAuthKeys() {
  val spec = KeyGenParameterSpec.Builder(Crypto.AUTH_KEY_ALIAS,
    KeyProperties.PURPOSE_SIGN or KeyProperties.PURPOSE_VERIFY)
    .setAlgorithmParameterSpec(ECGenParameterSpec("secp256r1"))
    .setDigests(KeyProperties.DIGEST_SHA256)
    .setCertificateSubject(X500Principal("CN=Acme Authenticator App"))
    .setCertificateSerialNumber(BigInteger.valueOf(Crypto.CERT_SERIAL))
    .setCertificateNotBefore(GregorianCalendar().time)
    .setCertificateNotAfter(GregorianCalendar().apply { add(GregorianCalendar.YEAR, 10) }.time)
    .build()
  KeyPairGenerator.getInstance(Crypto.KEY_ALGORITHM, Crypto.PROVIDER).apply {
    initialize(spec)
    generateKeyPair()
  }
}

fun generateUserKeys(keyAlias: String, userName: String, serial: Long) {
  val spec = KeyGenParameterSpec.Builder(keyAlias,
    KeyProperties.PURPOSE_SIGN or KeyProperties.PURPOSE_VERIFY)
    .setAlgorithmParameterSpec(ECGenParameterSpec("secp256r1"))
    .setDigests(KeyProperties.DIGEST_SHA256)
    .setCertificateSubject(X500Principal("CN=$userName"))
    .setCertificateSerialNumber(BigInteger.valueOf(serial))
    .setCertificateNotBefore(GregorianCalendar().time)
    .setCertificateNotAfter(GregorianCalendar().apply { add(GregorianCalendar.YEAR, 5) }.time)
    .build()
  KeyPairGenerator.getInstance(Crypto.KEY_ALGORITHM, Crypto.PROVIDER).apply {
    initialize(spec)
    generateKeyPair()
  }
}

fun deleteUserKeys(alias: String) {
  val keyStore = KeyStore.getInstance(Crypto.PROVIDER).apply { load(null) }
  if (keyStore.containsAlias(alias)) {
    keyStore.deleteEntry(alias)
  }
}

fun getKey(keyAlias: String): KeyStore.PrivateKeyEntry? {
  val keyStore = KeyStore.getInstance(Crypto.PROVIDER).apply { load(null) }
  return keyStore.getEntry(keyAlias, null) as? KeyStore.PrivateKeyEntry
}

fun getSignObject(keyAlias: String): BiometricPrompt.CryptoObject {
  val signer = Signature.getInstance(Crypto.SIGN_ALGORITHM)
  signer.initSign(getKey(keyAlias)!!.privateKey)
  return BiometricPrompt.CryptoObject(signer)
}

fun signMsg(signer: Signature, message: String): String {
  val msg = message.toByteArray(Charsets.UTF_8)                 // transform into a byte array
  signer.update(msg)                                                      // supply the bytes to sign
  val sig = signer.sign()                                       // perform the signature
  return Base64.encodeToString(sig, Base64.DEFAULT)                       // return it as a Base64 string
}

fun signMessage(keyAlias: String, message: String): String {
  val signer = Signature.getInstance(Crypto.SIGN_ALGORITHM)
  signer.initSign(getKey(keyAlias)!!.privateKey)
  val msg = message.toByteArray(Charsets.UTF_8)                 // transform into a byte array
  signer.update(msg)                                                      // supply the bytes to sign
  val sig = signer.sign()                                       // perform the signature
  return Base64.encodeToString(sig, Base64.DEFAULT)                       // return it as a Base64 string
}

fun randomString(length: Int): String {
  val allowedChars = ('A'..'Z') + ('a'..'z') + ('0'..'9')
  return (1..length)
    .map { allowedChars.random() }
    .joinToString("")
}

fun getShared(context: Context, name: String): SharedPreferences {
  val master = MasterKeys.getOrCreate(MasterKeys.AES256_GCM_SPEC)
  val sharedPref = EncryptedSharedPreferences.create(name, master, context,
    EncryptedSharedPreferences.PrefKeyEncryptionScheme.AES256_SIV,
    EncryptedSharedPreferences.PrefValueEncryptionScheme.AES256_GCM)
  return sharedPref
}