package org.feup.apm.myauthenticator

import android.content.Intent
import android.os.Bundle
import android.util.Base64
import android.view.Menu
import android.view.MenuItem
import android.widget.Button
import android.widget.TextView
import androidx.appcompat.app.AppCompatActivity
import androidx.biometric.BiometricManager
import androidx.biometric.BiometricPrompt
import androidx.core.content.ContextCompat
import androidx.lifecycle.lifecycleScope
import com.journeyapps.barcodescanner.ScanContract
import com.journeyapps.barcodescanner.ScanOptions
import java.util.UUID
import kotlin.random.Random
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import net.sqlcipher.database.SQLiteDatabase

class MainActivity : AppCompatActivity() {
  private val btScan by lazy { findViewById<Button>(R.id.bt_scan) }
  private val tvContent by lazy { findViewById<TextView>(R.id.tv_read_content) }
  private val tvAppName by lazy { findViewById<TextView>(R.id.tv_app_name) }
  private val tvResponse by lazy { findViewById<TextView>(R.id.tv_response) }

  override fun onCreate(savedInstanceState: Bundle?) {
    super.onCreate(savedInstanceState)
    setContentView(R.layout.activity_main)
    val sharedPref = getShared(this, "authShared")           // open/create shared preferences file
    if (sharedPref.getString("dbPass", "") == "") {          // if empty fill it with default values
      val newDBPass = randomString(16)                       // ... and a random DB password for encryption
      sharedPref.edit()
        .putString("dbPass", newDBPass)
        .putString("IP", "192.168.0.1")
        .putInt("port", 80)
        .putString("name", "Mike Mont")
        .apply()
    }
    if (getKey(Crypto.AUTH_KEY_ALIAS) == null)               // In the first execution after installation
      generateAuthKeys()                                     // generate the app asymmetric key-pair
    SQLiteDatabase.loadLibs(this)
    btScan.setOnClickListener { doScan() }                   // the scan button listener
    dbLayer = DBLayer(applicationContext)
  }

  override fun onCreateOptionsMenu(menu: Menu): Boolean {
    menuInflater.inflate(R.menu.menu_main, menu)
    return true
  }

  override fun onOptionsItemSelected(item: MenuItem): Boolean {
    return when (item.itemId) {
      R.id.action_settings -> {
        startActivity(Intent(this, SettingsActivity::class.java))  // change the default settings ...
        true
      }
      R.id.action_certificate -> {                                              // debug: view the app certificate
        startActivity(Intent(this, CertActivity::class.java))
        true
      }
      else -> super.onOptionsItemSelected(item)
    }
  }

  private fun doScan() {              // Scan a QRcode and obtain the info in the scanCodeLauncher callback
    tvContent.text = ""
    tvAppName.text = ""
    tvResponse.text = ""
    val options = ScanOptions().apply {
      setDesiredBarcodeFormats(ScanOptions.QR_CODE)
      setBeepEnabled(false)
      setOrientationLocked(true)
    }
    scanCodeLauncher.launch(options)
  }

  private val scanCodeLauncher = registerForActivityResult(ScanContract()) {
    var result = ""
    val list: List<String>
    if (it != null && it.contents != null) {
      result = it.contents
    }
    if (result.isNotEmpty()) {
      list = result.split(":", limit = 4)
      if (list.isNotEmpty()) {
        if (list[0] == "signup" && list.size == 3)                    // A signup QRCode
          performSignup(list)
        else if (list[0] == "signin" && list.size == 4)
          performSignIn(list)                                         // A signin QRCode
        tvContent.text = result
      }
    }
  }

  private fun getBiometricSignPrompt(message: String, result: (String) -> Unit): BiometricPrompt {
    val biometricPrompt = BiometricPrompt(this, ContextCompat.getMainExecutor(this),
      object: BiometricPrompt.AuthenticationCallback() {
        override fun onAuthenticationSucceeded(result: BiometricPrompt.AuthenticationResult) {
          super.onAuthenticationSucceeded(result)
          val signer = result.cryptoObject!!.signature
          result(signMsg(signer!!, message))
        }
        override fun onAuthenticationFailed() {
          super.onAuthenticationFailed()
          result("")
        }
        override fun onAuthenticationError(errorCode: Int, errString: CharSequence) {
          super.onAuthenticationError(errorCode, errString)
          result("")
        }
      }
    )
    return biometricPrompt
  }

  private fun biometricPrompt(title: String): BiometricPrompt.PromptInfo {
    val prompt = BiometricPrompt.PromptInfo.Builder()
      .setTitle(title)
      .setAllowedAuthenticators(BiometricManager.Authenticators.BIOMETRIC_STRONG)
      .setNegativeButtonText("Cancel")
      .build()
    return prompt
  }

  private fun performSignup(values: List<String>) {         // values: 0-type, 1-appId, 2-appName
    var kName = ""
    var uname = ""
    var iPAddress = ""
    var port = 0
    var name = ""
    var authCert = ""
    var userCert = ""

    fun ready(signature: String) {
      if (signature.isNotEmpty()) {
        lifecycleScope.launch {
          var response = signupResponse(iPAddress, port, name, uname, authCert, userCert, signature)
          if (!response.startsWith("ok")) {
            response = "Invalid request\n$response"
            deleteUserKeys(kName)
          }
          else
            dbLayer.insertUser(values[1], uname, name, kName)     // take note on the encrypted database
          setResponse(response)
        }
      }
      else {
        deleteUserKeys(kName)
        tvAppName.text = ""
        tvResponse.text = getString(R.string.tv_sign_biometric)
      }
    }

    val user: MutableList<String> = dbLayer.getUserByAppId(values[1])
    if (user.size > 0)
      tvAppName.text = getString(R.string.tv_signup_error, values[2], user[2])
    else {
      tvAppName.text = getString(R.string.tv_signup, values[2])
      val sharedPref = getShared(this, "authShared")    // get values from shared preferences file
      uname = UUID.randomUUID().toString()                                            // generate a unique user name (uname)
      iPAddress = sharedPref.getString("IP", "")!!
      port = sharedPref.getInt("port", 0)
      name = sharedPref.getString("name", "")!!
      authCert = Base64.encodeToString(getKey(Crypto.AUTH_KEY_ALIAS)!!.certificate.encoded, Base64.DEFAULT)
      do {                                                                            // generate a key name for the user key-pair
        kName = randomString(8)                                                // ... starting with a letter
      } while (kName.first().isDigit())
      val serial = Random.nextLong(from = 1L, until = 9999999999L)              // a random serial number for the user certificate
      generateUserKeys(kName, name, serial)                                           // generate the user key-pair
      userCert = Base64.encodeToString(getKey(kName)!!.certificate.encoded, Base64.DEFAULT)

      val message = name + uname + authCert + userCert                         // string message to sign
      getBiometricSignPrompt(message, ::ready).authenticate(biometricPrompt("Sign Up"), getSignObject(Crypto.AUTH_KEY_ALIAS))
    }
  }

  private fun performSignIn(values: List<String>) {                    // values: 0-type, 1-appId, 2-appName, 3-challenge
    val user: MutableList<String> = dbLayer.getUserByAppId(values[1])  // user: 0-appid, 1-userid, 2-name, 3-keyalias
    var iPAddress = ""
    var port = 0

    fun ready(signature: String) {
      if (signature.isNotEmpty()) {
        val message = user[1] + signature                          // string message to send and sign by the app
        val sigB64 = signMessage(Crypto.AUTH_KEY_ALIAS, message)   // signature in Base64
        lifecycleScope.launch {
          var response = signinResponse(iPAddress, port, user[1], signature, sigB64)
          if (!response.startsWith("ok")) {
            response = "Invalid request (not signed in)\n$response"
          }
          setResponse(response)
        }
      }
      else {
        tvAppName.text = ""
        tvResponse.text = getString(R.string.tv_sign_biometric)
      }
    }

    if (user.size == 0)
      tvAppName.text = getString(R.string.tv_signin_error, values[2])
    else {
      tvAppName.text = getString(R.string.tv_signin, values[2])
      val sharedPref = getShared(this, "authShared")    // get values from shared preferences file
      iPAddress = sharedPref.getString("IP", "")!!
      port = sharedPref.getInt("port", 0)

      getBiometricSignPrompt(values[3], ::ready).authenticate(biometricPrompt("Sign In"), getSignObject(user[3]))
    }
  }

  private suspend fun setResponse(response: String) {
    withContext(Dispatchers.Main) {
      tvAppName.text = ""
      tvResponse.text = response
    }
  }
}